#include <stdio.h>
#include <math.h>
#include <hrc.h>
#include <token.h>

#ifdef SGI
#include <libgen.h>
#else
#include <unistd.h>
#endif

Model	*readhrc(name)
char	*name;
{ 
	FILE	*fp;
	char	*tmp;
	Model	*rootmodel;

	if (( fp = fopen ( name, "r" )) == NULL )
	{
		fprintf( stderr, "readhrc: trouble opening %s\n", name);
		return((Model *)NULL);
	}

	init_tokenizer();

	/* header crap */
	mand_token(fp,"HRCH",EXIT);
	mand_token(fp,":",EXIT);
	mand_token(fp,"Softimage",WARN);
	mand_token(fp,"4D",WARN);
	mand_token(fp,"Creative",WARN);
	mand_token(fp,"Environment",WARN);

	if (mand_token(fp,"v3.49",QUIET))
	{
#ifdef 0
		fprintf(stderr,"Unknown version (file is not v3.49)\n");
		fprintf(stderr,"Unexpected errors may occur\n");
#endif
	}
	
	rootmodel = new_model();
	parse_model(fp,rootmodel);

#ifdef DEBUG
	fprintf(stderr,"endreadhrc\n");
#endif

	fclose(fp);

	return( rootmodel );
}

int		parse_model( fp, model )
FILE	*fp;
Model	*model;
{
	char	*token;
	int		done;
	Model	*mod;
	Mesh	*mesh;
	Cluster	*clst;
	Material	*mtl;
	Texture	*txt;

	mand_token(fp,"model",EXIT);
	mand_token(fp,"{",EXIT);
	mand_token(fp,"name",EXIT);

	model->name = get_string(fp);
	
	mand_token(fp,"scaling",EXIT);
	mand_float3(fp,&(model->scale),"scale x,y,z",EXIT);

	mand_token(fp,"rotation",EXIT);
	mand_float3(fp,&(model->rotate),"rotate x,y,z",EXIT);

	mand_token(fp,"translation",EXIT);
	mand_float3(fp,&(model->translate),"translate x,y,z",EXIT);

	/* onward */

	done = 0;
	while ( !done )
	{
		if ( opt_token( fp,"model" ))
		{
			mod = new_model();
			ins_list(&(model->childlist), mod);
			parse_model( fp, mod );
		}
		else 
		if ( opt_token(fp,"mesh") )
		{
			mesh = new_mesh();
			model->mesh = mesh;
			parse_mesh( fp, mesh );
		}
		else
		if ( opt_token(fp,"texture") )
		{
			txt = new_texture();
			ins_list(&(model->txtlist), txt);

			parse_textures( fp, txt );
		}
		else
		if ( opt_token(fp,"material") )
		{
			mtl = new_material();
			ins_list(&(model->mtllist), mtl);

			parse_material( fp, mtl );
		}
		else
		if ( opt_token(fp,"DK_clusters") )
		{
			parse_clusters( fp, &(model->clstlist));
		}
		else
		if ( opt_token(fp,"}") )
		{
			mand_token( fp, "}" );
			done = 1;
		}
		else
		{
			token = get_token(fp);
			fprintf(stderr,"Unexpected token %s\n",token);
			exit(-1);
		}
	}

	/* end model */

#ifdef DEBUG
	fprintf(stderr,"endmodel\n");
#endif

}

parse_mesh( fp, mesh )
FILE	*fp;
Mesh	*mesh;
{
	Vertex	*vertlist;
	Polygon	*polylist;
	Node	*nodelist;
	int		i,j,tmpint;

	mand_token(fp,"mesh",EXIT);
	mand_token(fp,"{",EXIT);

	if ( opt_token(fp,"flag"))
	{
		mand_token(fp,"flag",EXIT);
		mand_token(fp,"(",EXIT);

		mesh->flag = mand_symbol(fp,WARN);
		if ( opt_symbol(fp) )
			mesh->flag |= mand_symbol(fp,WARN);

		if ( opt_symbol(fp) )
			mesh->flag |= mand_symbol(fp,WARN);

		mand_token(fp,")",EXIT);
	}

	mand_token(fp,"discontinuity",EXIT);
	mand_float(fp,&(mesh->discontinuity),"discontinuity",EXIT);

	/* vertices */

	mand_token(fp,"vertices",EXIT);
	mand_int(fp,&(mesh->nverts),"Number of vertices",EXIT);

	mesh->vertlist = vertlist = 
		(Vertex *)malloc( mesh->nverts * sizeof(Vertex));

	mesh->bbox_min[0] = mesh->bbox_min[1] = mesh->bbox_min[2] =  HUGE;
	mesh->bbox_max[0] = mesh->bbox_max[1] = mesh->bbox_max[2] = -HUGE;

	mand_token(fp,"{",EXIT);
	for (i=0; i<mesh->nverts; i++)
	{
		mand_token(fp,"[",EXIT);
		mand_int(fp,&tmpint,"vertex number",EXIT);
		mand_token(fp,"]",EXIT);

		mand_token(fp,"position",EXIT);
		mand_float3(fp,&(vertlist[i].p),"position x,y,z",EXIT);

		if ( mesh->bbox_min[0] > vertlist[i].p[0] )
			mesh->bbox_min[0] = vertlist[i].p[0];

		if ( mesh->bbox_min[1] > vertlist[i].p[1] )
			mesh->bbox_min[1] = vertlist[i].p[1];

		if ( mesh->bbox_min[2] > vertlist[i].p[2] )
			mesh->bbox_min[2] = vertlist[i].p[2];

		if ( mesh->bbox_max[0] < vertlist[i].p[0] )
			mesh->bbox_max[0] = vertlist[i].p[0];

		if ( mesh->bbox_max[1] < vertlist[i].p[1] )
			mesh->bbox_max[1] = vertlist[i].p[1];

		if ( mesh->bbox_max[2] < vertlist[i].p[2] )
			mesh->bbox_max[2] = vertlist[i].p[2];

		vertlist[i].flag = 0;

		if ( opt_token( fp,"flag" ))
		{
			mand_token(fp,"flag",EXIT);
			mand_token(fp,"(",EXIT);
			vertlist[i].flag = mand_symbol(fp,WARN);
			mand_token(fp,")",EXIT);
		}

		vertlist[i].xflag = 1;
		vertlist[i].tmp = 0;
		vertlist[i].w = 0;
		vertlist[i].d[0] = 0;
		vertlist[i].d[1] = 0;
		vertlist[i].d[2] = 0;
		vertlist[i].dm = 0;
		
		vertlist[i].elist.ptr = (void *)0;
		vertlist[i].elist.next = (Llist *)0;
	}
	mand_token(fp,"}",EXIT);
	/* end vertices */

	/* polygons */
	mand_token(fp,"polygons",EXIT);
	mand_int(fp,&(mesh->npolys),"Number of polygons",EXIT);

	mesh->polylist = polylist = 
		(Polygon *)malloc( mesh->npolys * sizeof(Polygon));

	mand_token(fp,"{",EXIT);
	for (i=0; i<mesh->npolys; i++)
	{
		mand_token(fp,"[",EXIT);
		mand_int(fp,&tmpint,"polygon number",EXIT);
		mand_token(fp,"]",EXIT);

		mand_token(fp,"nodes",EXIT);
		mand_int(fp,&(polylist[i].nodes),"number of nodes",EXIT);

		polylist[i].nodelist = nodelist = (Node *)malloc(polylist[i].nodes * 
			sizeof(Node));

		/* begin nodes */
		mand_token(fp,"{",EXIT);
		for (j=0; j<polylist[i].nodes; j++)
		{
			mand_token(fp,"[",EXIT);
			mand_int(fp,&tmpint,"node number",EXIT);
			mand_token(fp,"]",EXIT);

			mand_token(fp,"vertex",EXIT);
			mand_int(fp,&(nodelist[j].vnum),"vnum number",EXIT);

			mand_token(fp,"normal",EXIT);
			mand_float3(fp,&(nodelist[j].n),"normal x,y,z",EXIT);

			mand_token(fp,"uvTexture",EXIT);
			mand_float2(fp,&(nodelist[j].uv),"uv x,y",EXIT);
		}
		mand_token(fp,"}",EXIT);
		/* end nodes */

		/* poly material */
		mand_token(fp,"material",EXIT);
		mand_int(fp,&(polylist[i].mtlnum),"material number",EXIT);

		/* optional flags */
		if ( opt_token( fp,"flag" ))
		{
			mand_token(fp,"flag",EXIT);
			mand_token(fp,"(",EXIT);
			vertlist[i].flag = mand_symbol(fp,WARN);
			mand_token(fp,")",EXIT);
		}
	}
	mand_token(fp,"}",EXIT);
	/* end polygons */

	if (opt_token( fp,"edges" ))
		parse_edges(fp,mesh);

	mand_token(fp,"}",EXIT);
	/* end mesh */

#ifdef DEBUG
	fprintf(stderr,"endmesh\n");
#endif
}

parse_edges(fp,mesh)
FILE    *fp;
Mesh    *mesh;
{
	int	i,tmpint;
	Edge	*edgelist;

	/* begin edges */
	mand_token(fp,"edges",EXIT);
	mand_int(fp,&(mesh->nedges),"number of edges",EXIT);
	mesh->edgelist = edgelist = 
		(Edge *)malloc( mesh->nedges * sizeof(Edge));

	mand_token(fp,"{",EXIT);
	for (i=0; i<mesh->nedges; i++)
	{
		mand_token(fp,"[",EXIT);
		mand_int(fp,&tmpint,"edge number",EXIT);
		mand_token(fp,"]",EXIT);

		mand_token(fp,"vertices",EXIT);
		mand_int(fp,&(edgelist[i].vptr0),"vnum",EXIT);
		mand_int(fp,&(edgelist[i].vptr1),"vnum",EXIT);

		if ( opt_token( fp,"flag" ))
		{
			mand_token(fp,"flag",EXIT);
			mand_token(fp,"(");
			edgelist[i].flag = mand_symbol(fp,WARN);
			mand_token(fp,")");
		}
	}
	mand_token(fp,"}",EXIT);
	/* end edges */
}

int	parse_textures( fp, txt )
FILE	*fp;
Texture	*txt;
{
	int	tmpint;
	char	*token;

	mand_token(fp,"texture",EXIT);
	mand_token(fp,"[",EXIT);
	mand_int(fp,&tmpint,"texture number",EXIT);
	mand_token(fp,"]",EXIT);

	mand_token(fp,"{",EXIT);

	mand_token(fp,"name",EXIT);
	txt->picname = get_string(fp);
	
	mand_token(fp,"glbname",EXIT);
	txt->txtname = get_string(fp);

	mand_token(fp,"anim",EXIT);
	txt->anim = mand_symbol(fp,WARN);

	mand_token(fp,"method",EXIT);
	txt->method = mand_symbol(fp,WARN);
	
	mand_token(fp,"repeat",EXIT);
	mand_float2(fp,&(txt->repeat),"repeat x,y",EXIT);

	mand_token(fp,"scaling",EXIT);
	mand_float2(fp,&(txt->scaling),"scaling x,y",EXIT);

	mand_token(fp,"offset",EXIT);
	mand_float2(fp,&(txt->offset),"offset x,y",EXIT);

	if ( opt_token(fp,"pixelInterp",EXIT))
	{
		mand_token(fp,"pixelInterp",EXIT);
		txt->interp = 1;
	}

	if ( opt_token(fp,"uvswap",EXIT))
	{
		mand_token(fp,"uvswap",EXIT);
		txt->uvswap = 1;
	}

	mand_token(fp,"effect",EXIT);
	txt->effect = mand_symbol(fp,WARN);

	if (opt_token(fp,"mask"))
	{
		mand_token(fp,"mask",EXIT);
		txt->mask = mand_symbol(fp,WARN);
	}

	mand_token(fp,"blending",EXIT);
	mand_float(fp,&(txt->blending),"blending",EXIT);

	mand_token(fp,"ambient",EXIT);
	mand_float(fp,&(txt->amb),"ambient",EXIT);

	mand_token(fp,"diffuse",EXIT);
	mand_float(fp,&(txt->diff),"diffuse",EXIT);

	mand_token(fp,"specular",EXIT);
	mand_float(fp,&(txt->spec),"specular",EXIT);

	mand_token(fp,"reflect",EXIT);
	mand_float(fp,&(txt->rfl),"reflect",EXIT);

	mand_token(fp,"transp",EXIT);
	mand_float(fp,&(txt->trans),"transp",EXIT);

	mand_token(fp,"roughness",EXIT);
	mand_float(fp,&(txt->roughness),"roughness",EXIT);

	mand_token(fp,"reflMap",EXIT);
	mand_float(fp,&(txt->rflmap),"reflMap",EXIT);

	mand_token(fp,"rotation",EXIT);
	mand_float(fp,&(txt->rotation),"rotation",EXIT);

	if ( opt_token(fp,"cropping",EXIT))
	{
		mand_token(fp,"cropping",EXIT);
		mand_int(fp,&(txt->lleftx),"lleftx",EXIT);
		mand_int(fp,&(txt->llefty),"llefty",EXIT);
		mand_int(fp,&(txt->urightx),"urightx",EXIT);
		mand_int(fp,&(txt->urighty),"urighty",EXIT);
	}

	mand_token(fp,"txtsup_rot",EXIT);
	mand_float3(fp,&(txt->tsup_rot),"txtsup_rot",EXIT);

	mand_token(fp,"txtsup_trans",EXIT);
	mand_float3(fp,&(txt->tsup_trans),"txtsup_trans",EXIT);

	mand_token(fp,"txtsup_scal",EXIT);
	mand_float3(fp,&(txt->tsup_scale),"txtsup_scal",EXIT);

	mand_token(fp,"}",EXIT);

#ifdef DEBUG
	fprintf(stderr,"endtxt\n");
#endif

	return(1);
}

int	parse_material( fp, mtl )
FILE	*fp;
Material	*mtl;
{
	int	done;
	Texture	*txt;

	mand_token(fp,"material",EXIT);
	mand_token(fp,"[",EXIT);
	mand_int(fp,&(mtl->mtlnum),"material number",EXIT);
	mand_token(fp,"]",EXIT);

	mand_token(fp,"{",EXIT);

	mand_token(fp,"name",EXIT);
	mtl->name = get_string(fp);

	mand_token(fp,"type",EXIT);
	mtl->type = mand_symbol(fp,WARN);

	mand_token(fp,"ambient",EXIT);
	mand_float3(fp,&(mtl->amb),"ambient",EXIT);

	mand_token(fp,"diffuse",EXIT);
	mand_float3(fp,&(mtl->diff),"diffuse",EXIT);

	mand_token(fp,"specular",EXIT);
	mand_float3(fp,&(mtl->spec),"specular",EXIT);

	mand_token(fp,"exponent",EXIT);
	mand_float(fp,&(mtl->exp),"exponent",EXIT);

	mand_token(fp,"reflectivity",EXIT);
	mand_float(fp,&(mtl->rfl),"reflectivity",EXIT);

	mand_token(fp,"transparency",EXIT);
	mand_float(fp,&(mtl->trans),"transparency",EXIT);

	mand_token(fp,"refracIndex",EXIT);
	mand_float(fp,&(mtl->refract),"refracIndex",EXIT);

	mand_token(fp,"glow",EXIT);
	mand_int(fp,&(mtl->glow),"glow",EXIT);

	mand_token(fp,"coc",EXIT);
	mand_float(fp,&(mtl->coc),"coc",EXIT);


	done = 0;
	while ( !done )
	{
		if ( opt_token(fp,"texture"))
		{
			txt = new_texture();
			ins_list(&(mtl->txtlist), txt);

			parse_textures( fp, txt );
		}
		else
		if ( opt_token(fp,"}"))
			done = 1;
	}

	mand_token(fp,"}",EXIT);

#ifdef DEBUG
	fprintf(stderr,"endmtl\n");
#endif

	return(1);
}

int	parse_clusters( fp, clstlist )
FILE	*fp;
Llist	*clstlist;
{
	int	done,i;
	Cluster	*clst;

	mand_token(fp,"DK_clusters",EXIT);
	mand_token(fp,"{",EXIT);

	done = 0;
	while ( !done )
	{
		mand_token(fp,"DK_cluster_name",EXIT);

		clst = new_cluster();
		ins_list(clstlist, clst);

		clst->name = get_string(fp);

#ifdef DBUG
		fprintf(stdout,"Cluster %s read..",clst->name);
#endif
		mand_token(fp,"DK_cluster_ncvs",EXIT);
		mand_int(fp,&(clst->ncvs),"ncvs",EXIT);

		mand_token(fp,"DK_cluster_vertices",EXIT);

		clst->vlist = (int *)malloc( sizeof(int) * clst->ncvs );

		mand_token(fp,"{",EXIT);
		for (i=0; i<clst->ncvs; i++)
		{
			mand_int(fp,&(clst->vlist[i]),"vertex no.",EXIT);
		}
		mand_token(fp,"}",EXIT);

		mand_token(fp,"DK_cluster_state",EXIT);
		mand_int(fp,&(clst->state),"DK_cluster_state",EXIT);

		if (opt_token( fp,"}"))
			done = 1;

#ifdef DBUG
		fprintf(stdout,"cluster state %d\n",clst->state);
#endif
	}
	mand_token(fp,"}",EXIT);

#ifdef DEBUG
	fprintf(stderr,"endclst\n");
#endif

	return(1);
}

ins_list(ll, item)
Llist	*ll;
void	*item;
{
	int	done;

	if ( ll->ptr == (void *)NULL )
		ll->ptr = item;
	else
	{
		done = 0;
		while (!done) 
		{
			if ( ll->next == (void *)NULL )
			{
				ll->next = (Llist *)malloc( sizeof(Llist) );
				ll->next->next = (Llist *)NULL;
				ll->next->ptr = item;
				done = 1;
			}
			else
				ll = ll->next;
		}
	}
	
}

writehrc( name, mod )
char	*name;
Model	*mod;
{
	FILE	*fp;

	if (( fp = fopen ( name, "w" )) == NULL )
	{
		fprintf( stderr, "writehrc: trouble opening %s\n", name);
		exit(-1);
	}

	fprintf(fp,"HRCH: Softimage 4D Creative Environment v3.49\n");

	print_model(fp,mod);
	fclose(fp);
}

char    *spaces="                                                             ";
#define	SP 61
int		rdepth=0;

print_model( fp, mod )
FILE	*fp;
Model	*mod;
{
	int	done,i,j;
	Llist	*child,*mtll,*txtl,*clstl;

	fprintf(fp,"%smodel\n",&(spaces[SP-rdepth]));
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

	rdepth+=2;

	if ( mod->name == (char *)NULL )
		fprintf(fp,"%sname\t%s",&(spaces[SP-rdepth]),"\"NoNaMe\"");
	else
		fprintf(fp,"%sname\t%s",&(spaces[SP-rdepth]),mod->name);

	fprintf(fp,"%sscaling\t",&(spaces[SP-rdepth])); 
	print3f(fp,"%f %f %f\n",mod->scale);

	fprintf(fp,"%srotation\t",&(spaces[SP-rdepth])); 
	print3f(fp,"%f %f %f\n",mod->rotate);

	fprintf(fp,"%stranslation\t",&(spaces[SP-rdepth])); 
	print3f(fp,"%f %f %f\n",mod->translate);

	done = 0;
	child = &(mod->childlist);
	while ( !done )
	{
		if ( child->ptr != (void *)NULL )
			print_model(fp,child->ptr);

		if ( child->next == (Llist *)NULL )
			done = 1;
		else
			child = child->next;
	}

	if ( mod->mesh != (Mesh *)NULL )
		print_mesh( fp, mod->mesh );

	done = 0;
	mtll = &(mod->mtllist);
	while ( !done )
	{
		if ( mtll->ptr != (void *)NULL )
			print_material(fp,mtll->ptr);

		if ( mtll->next == (Llist *)NULL )
			done = 1;
		else
			mtll = mtll->next;
	}

	done = 0;
	txtl = &(mod->txtlist);
	while ( !done )
	{
		if ( txtl->ptr != (void *)NULL )
			print_texture(fp,txtl->ptr);

		if ( txtl->next == (Llist *)NULL )
			done = 1;
		else
			txtl = txtl->next;
	}

	print_clusters(fp,&(mod->clstlist));

	rdepth-=2;
	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
}

print_mesh(fp,mesh)
FILE	*fp;
Mesh	*mesh;
{
	int	i,j,done;

	fprintf(fp,"%smesh\n",&(spaces[SP-rdepth])); 
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

	rdepth += 2;

	fprintf(fp,"%sflag\t",&(spaces[SP-rdepth])); 
	fprintf(fp,"( %s )\n",get_symname(mesh->flag));

	fprintf(fp,"%sdiscontinuity\t",&(spaces[SP-rdepth])); 
	fprintf(fp,"%f\n",mesh->discontinuity);


	/* vertices */
	fprintf(fp,"%svertices\t",&(spaces[SP-rdepth])); 
	fprintf(fp,"%d\n",mesh->nverts);
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

	rdepth += 2;
	for (i=0; i<mesh->nverts; i++)
	{
		fprintf(fp,"%s[%d] position\t",&(spaces[SP-rdepth]),i); 
		print3f(fp,"%f %f %f\n",mesh->vertlist[i].p);
		if ( mesh->vertlist[i].flag != 0 )
			fprintf(fp,"%s    flag ( %s )\n",&(spaces[SP-rdepth]),
				get_symname(mesh->vertlist[i].flag)); 
	}
	rdepth -= 2;
	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));

	/* polygons */
	fprintf(fp,"%spolygons\t",&(spaces[SP-rdepth])); 
	fprintf(fp,"%d\n",mesh->npolys);
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));
	rdepth += 2;
	for (i=0; i<mesh->npolys; i++)
	{
		fprintf(fp,"%s[%d] nodes %d\n",&(spaces[SP-rdepth]),i,
			mesh->polylist[i].nodes); 
		rdepth += 4;
		fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));
		rdepth += 2;

		for (j=0; j<mesh->polylist[i].nodes; j++)
		{
			fprintf(fp,"%s[%d] vertex\t%d\n",&(spaces[SP-rdepth]),j,
				mesh->polylist[i].nodelist[j].vnum); 

			fprintf(fp,"%snormal\t",&(spaces[SP-rdepth]));
			print3f(fp,"%f %f %f\n",mesh->polylist[i].nodelist[j].n); 

			fprintf(fp,"%suvTexture\t",&(spaces[SP-rdepth]));
			fprintf(fp,"%f %f\n",mesh->polylist[i].nodelist[j].uv[0],
				mesh->polylist[i].nodelist[j].uv[1]); 

		}
		rdepth -= 2;
		fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
		fprintf(fp,"%smaterial\t%d\n",&(spaces[SP-rdepth]),
			mesh->polylist[i].mtlnum);
		rdepth -= 4;
	}

	rdepth -= 2;
	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));

	/* edges */
	if ( mesh->nedges != 0 )
	{
		fprintf(fp,"%sedges\t",&(spaces[SP-rdepth])); 
		fprintf(fp,"%d\n",mesh->nedges);
		fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));
		rdepth += 2;
		for (i=0; i<mesh->nedges; i++)
		{
			fprintf(fp,"%s[%d] vertices\t%d\t%d\n",&(spaces[SP-rdepth]),i+1,
				mesh->edgelist[i].vptr0,mesh->edgelist[i].vptr1); 
			if ( mesh->edgelist[i].flag )
				fprintf(fp,"%sflag\t( %s )\n",&(spaces[SP-rdepth]), 
					get_symname(mesh->edgelist[i].flag)); 
		}
		rdepth -= 2;
		fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
	}

	rdepth -= 2;
	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
}

print_clusters(fp,clstlist)
FILE	*fp;
Llist	*clstlist;
{

	Cluster	*clst;
	Llist	*clstl;
	int	done,i;

	clst = clstlist->ptr;

	if ( clst != (void *)NULL )
	{
		fprintf(fp,"%sDK_clusters\t",&(spaces[SP-rdepth])); 
		fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

		rdepth += 2;

		done = 0;
		while ( !done )
		{
			fprintf(fp,"%sDK_cluster_name\t%s\n",&(spaces[SP-rdepth]),
				clst->name); 
			fprintf(fp,"%sDK_cluster_ncvs\t%d\n",&(spaces[SP-rdepth]),
				clst->ncvs); 
			fprintf(fp,"%sDK_cluster_vertices\t\n",&(spaces[SP-rdepth])); 
			fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

			rdepth += 2;
			for (i=0; i<clst->ncvs; i++)
				fprintf(fp,"%s%d\n",&(spaces[SP-rdepth]),clst->vlist[i]); 
			rdepth -= 2;

			fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
			fprintf(fp,"%sDK_cluster_state\t%d\n\n",&(spaces[SP-rdepth]),
				clst->state); 

			/* next cluster */
			if ( clstlist->next != (void *)NULL )
			{
				clstlist = clstlist->next;
				clst = clstlist->ptr;
			}
			else
				done = 1;
		}
		rdepth -= 2;
		fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
	}
}

print_texture(fp,txt)
FILE	*fp;
Texture	*txt;
{

	fprintf(fp,"%stexture\t[%d]\n",&(spaces[SP-rdepth]),txt->txtnum); 
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));

	rdepth += 2;

	fprintf(fp,"%sname\t%s\n",&(spaces[SP-rdepth]),txt->picname);
	fprintf(fp,"%sglbname\t%s\n",&(spaces[SP-rdepth]),txt->txtname);
	fprintf(fp,"%sanim\t%s\n",&(spaces[SP-rdepth]),get_symname(txt->anim));
	fprintf(fp,"%smethod\t%s\n",&(spaces[SP-rdepth]),get_symname(txt->method));

	fprintf(fp,"%srepeat\t%f %f\n",&(spaces[SP-rdepth]),txt->repeat[0],
		txt->repeat[1]);

	fprintf(fp,"%sscaling\t%f %f\n",&(spaces[SP-rdepth]),txt->scaling[0],
		txt->scaling[1]);

	fprintf(fp,"%soffset\t%f %f\n",&(spaces[SP-rdepth]),txt->offset[0],
		txt->offset[1]);

	if ( txt->interp )
		fprintf(fp,"%spixelInterp\n",&(spaces[SP-rdepth]));

	fprintf(fp,"%seffect\t%s\n",&(spaces[SP-rdepth]),get_symname(txt->effect));

	fprintf(fp,"%sblending\t%f\n",&(spaces[SP-rdepth]),txt->blending);
	fprintf(fp,"%sambient\t%f\n",&(spaces[SP-rdepth]),txt->amb);
	fprintf(fp,"%sdiffuse\t%f\n",&(spaces[SP-rdepth]),txt->diff);
	fprintf(fp,"%sspecular\t%f\n",&(spaces[SP-rdepth]),txt->spec);
	fprintf(fp,"%sreflect\t%f\n",&(spaces[SP-rdepth]),txt->rfl);
	fprintf(fp,"%stransp\t%f\n",&(spaces[SP-rdepth]),txt->trans);
	fprintf(fp,"%sroughness\t%f\n",&(spaces[SP-rdepth]),txt->roughness);
	fprintf(fp,"%sreflMap\t%f\n",&(spaces[SP-rdepth]),txt->rflmap);
	fprintf(fp,"%srotation\t%f\n",&(spaces[SP-rdepth]),txt->rotation);

	fprintf(fp,"%stxtsup_rot\t%f %f %f\n",&(spaces[SP-rdepth]),
		txt->tsup_rot[0],txt->tsup_rot[1],txt->tsup_rot[2]);

	fprintf(fp,"%stxtsup_trans\t%f %f %f\n",&(spaces[SP-rdepth]),
		txt->tsup_trans[0],txt->tsup_trans[1],txt->tsup_trans[2]);

	fprintf(fp,"%stxtsup_scal\t%f %f %f\n",&(spaces[SP-rdepth]),
		txt->tsup_scale[0],txt->tsup_scale[1],txt->tsup_scale[2]);

	rdepth -= 2;

	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
}

print_material(fp,mtl)
FILE	*fp;
Material	*mtl;
{
	int	done;
	Llist	*txtl;

	fprintf(fp,"%smaterial\t[%d]\n",&(spaces[SP-rdepth]),mtl->mtlnum); 
	fprintf(fp,"%s{\n",&(spaces[SP-rdepth]));
	rdepth += 2;

	fprintf(fp,"%sname\t%s\n",&(spaces[SP-rdepth]),mtl->name); 
	fprintf(fp,"%stype\t%s\n",&(spaces[SP-rdepth]),get_symname(mtl->type));

	fprintf(fp,"%sambient\t%f %f %f\n",&(spaces[SP-rdepth]),
		mtl->amb[0],mtl->amb[1],mtl->amb[2]);

	fprintf(fp,"%sdiffuse\t%f %f %f\n",&(spaces[SP-rdepth]),
		mtl->diff[0],mtl->diff[1],mtl->diff[2]);

	fprintf(fp,"%sspecular\t%f %f %f\n",&(spaces[SP-rdepth]),
		mtl->spec[0],mtl->spec[1],mtl->spec[2]);

	fprintf(fp,"%sexponent\t%f\n",&(spaces[SP-rdepth]),mtl->exp);
	fprintf(fp,"%srelectivity\t%f\n",&(spaces[SP-rdepth]),mtl->rfl);
	fprintf(fp,"%stransparency\t%f\n",&(spaces[SP-rdepth]),mtl->trans);
	fprintf(fp,"%srefracIndex\t%f\n",&(spaces[SP-rdepth]),mtl->refract);
	fprintf(fp,"%sglow\t%d\n",&(spaces[SP-rdepth]),mtl->glow);
	fprintf(fp,"%scoc\t%f\n",&(spaces[SP-rdepth]),mtl->coc);

	done = 0;
	txtl = &(mtl->txtlist);
	while ( !done )
	{
		if ( txtl->ptr != (void *)NULL )
			print_texture(fp,txtl->ptr);

		if ( txtl->next == (Llist *)NULL )
			done = 1;
		else
			txtl = txtl->next;
	}

	rdepth -= 2;
	fprintf(fp,"%s}\n",&(spaces[SP-rdepth]));
}

print3f(fp,fmt,p)
FILE	*fp;
char	*fmt;
float	p[];
{
	fprintf(fp,fmt,p[0],p[1],p[2]);
}
