#define TRUE            1
#define FALSE           0

#define FACETYPE        0
#define HOLETYPE        1

#ifndef MAXFILENAME
#define MAXFILENAME 256
#endif

/* max points per poly */
#define MAXPPPOLY 1024

/* num of elem malloc each time */
#define NUMMALLOC	256

typedef struct Llist {
	void	*ptr;
	struct	Llist	*next;
} Llist;

typedef struct Vertex {
	float	pp[3];	/* pristine */
	float	pn[3];	/* ditto-normal */

	float	op[3];	/* "old" p */
	float	on[3];	/* "old" normal */

	float	p[3];	/* p (xyz) */
	float	n[3];	/* normal */

	float	d[3];	/* delta */
	float	dm;		/* delta magnitude */

	float	w;		/* weight */

	int		flag;
	int		xflag;	/* xform flag */
	int		tmp;	/* clobbered by pointnormal() */

	Llist	elist;	/* edge list */
} Vertex;

typedef struct Node {
	int		vnum;
	float	uv[2];
	float	n[3];
	int		flag;
} Node;

typedef struct Polygon {
	int		nodes;
	float	n[3];
	int		mtlnum;
	int		flag;
	Node	*nodelist;
} Polygon;

typedef struct Edge {
	int	vptr0;
	int	vptr1;
	float	length;
	float	d[3];
	int	flag;
} Edge;

typedef struct Cluster {
	char	*name;
	int		ncvs;
	int		state;
	int		*vlist;
} Cluster;

typedef struct Texture {
	int		txtnum;
	char	*picname;
	char	*txtname;
	int		anim;
	int		method;
	float	repeat[2];
	float	scaling[2];
	float	offset[2];
	int		interp;		/* pixelInterp on/off */
	int		uvswap;		/* swap row/column */
	int		effect;		/* intensity etc */
	int		mask;		/* intensity etc */
	float	blending;
	float	amb;
	float	diff;
	float	spec;
	float	rfl;
	float	trans;
	float	roughness;
	float	rflmap;
	float	rotation;
	float	lleftx,llefty,urightx,urighty;
	float	tsup_rot[3];
	float	tsup_trans[3];
	float	tsup_scale[3];
} Texture;

typedef	union Lptr {
		struct	Model	*mod;
		struct	Texture	*txt;
		struct	Material	*mtl;
		struct	Cluster	*clst;
} Lptr;

typedef struct Material {
	int		mtlnum;
	char	*name;
	int		type;
	float	amb[3];
	float	diff[3];
	float	spec[3];
	float	exp;
	float	rfl;
	float	trans;
	float	refract;
	int		glow;
	float	coc;
	Llist	txtlist;
} Material;

typedef struct Mesh {
	char	*name;

	int		flag;
	float	discontinuity;

	Vertex	*vertlist;
	Edge	*edgelist;
	Polygon	*polylist;
	Cluster	*clstlist;

	int		nverts;
	int		nedges;
	int		npolys;
	int		nclsts;

	/* internal use only */
	int		maxverts;
	int		maxpolys;
	int		maxedges;
	int		maxclsts;

	float	bbox_min[3];
	float	bbox_max[3];
} Mesh;

typedef struct Model {
	char	*name;
	int		type;	/* poly,spline, patch etc */
	float	scale[3];
	float	rotate[3];
	float	translate[3];
	Mesh	*mesh;
	Llist	childlist;
	Llist	mtllist;
	Llist	txtlist;
	Llist	clstlist;
} Model;

extern	char	*unget_token_ptr,*token_ptr;
extern	char	token_buffer[];
extern	int	linecounter;

/* symbol table stuff */
typedef struct Symtab {
	char	*name;
	int		num;
} Symtab;

int		parse_model();
int		parse_mesh();
int		parse_material();
int		parse_clusters();

Model	*readhrc();
Model	*new_model();
Mesh	*new_mesh();
Cluster	*new_cluster();
Texture	*new_texture();
Material	*new_material();
int		new_vertex();
int		new_poly();
Node	*new_nodes();
