#include <stdio.h>
#include <hrc.h>

/* create and init new model */

Model	*new_model()
{
	Model	*mod;

	mod = (Model *)malloc(sizeof(Model));

	mod->name = (char *)NULL;
	mod->type = 0;

	mod->scale[0] = 1.0; mod->scale[1] = 1.0; mod->scale[2] = 1.0;
	mod->rotate[0] = 0; mod->rotate[1] = 0; mod->rotate[2] = 0; 
	mod->translate[0] = 0; mod->translate[1] = 0; mod->translate[2] = 0; 

	mod->childlist.ptr = (Model *)NULL;
	mod->childlist.next = (Llist *)NULL;

	mod->mtllist.ptr = (Material *)NULL;
	mod->mtllist.next = (Llist *)NULL;

	mod->txtlist.ptr = (Texture *)NULL;
	mod->txtlist.next = (Llist *)NULL;

	mod->clstlist.ptr = (Cluster *)NULL;
	mod->clstlist.next = (Llist *)NULL;

	mod->mesh = (Mesh *)NULL;

	return( mod );
}

/* create and init new mesh */

Mesh	*new_mesh()
{
	Mesh	*me;

	me = (Mesh *)malloc(sizeof(Mesh));

	me->name = (char *)NULL;
	me->flag = 0;
	me->discontinuity = 0.0;
	me->nverts = 0;
	me->nedges = 0;
	me->npolys = 0;
	me->nclsts = 0;

	/* internal-for malloc purposes */
	me->maxverts = 0;
	me->maxpolys = 0;
	me->maxedges = 0;
	me->maxclsts = 0;

	me->vertlist = (Vertex *)NULL;
	me->edgelist = (Edge *)NULL;
	me->polylist = (Polygon *)NULL;
	me->clstlist = (Cluster *)NULL;

	return( me );
}

Texture	*new_texture()
{
	Texture	*txt;

	txt = (Texture *)malloc(sizeof(Texture));

	txt->txtnum = 0;
	txt->picname = (char *)NULL;
	txt->txtname = (char *)NULL;
	txt->anim = 0;
	txt->method = 0;
	txt->repeat[0] = txt->repeat[1] = 0;
	txt->scaling[0] = txt->scaling[1] = 0;
	txt->offset[0] = txt->offset[1] = 0;
	txt->interp = 0;
	txt->effect = 0;
	txt->uvswap = 0;
	txt->blending = 0;
	txt->amb = 0;
	txt->diff = 0;
	txt->spec = 0;
	txt->rfl = 0;
	txt->trans = 0;
	txt->roughness = 0;
	txt->rflmap = 0;
	txt->rotation = 0;
	txt->tsup_rot[0] = txt->tsup_rot[1] = txt->tsup_rot[2] = 0;
	txt->tsup_trans[0] = txt->tsup_trans[1] = txt->tsup_trans[2] = 0;
	txt->tsup_scale[0] = txt->tsup_scale[1] = txt->tsup_scale[2] = 0;

	return(txt);
}

Material	*new_material()
{
	Material	*mtl;

	mtl = (Material *)malloc(sizeof(Material));

	mtl->mtlnum = 0;
	mtl->name = (char *)NULL;
	mtl->type = 0;
	mtl->amb[0] = mtl->amb[1] = mtl->amb[2] = 0;
	mtl->diff[0] = mtl->diff[1] = mtl->diff[2] = 0;
	mtl->spec[0] = mtl->spec[1] = mtl->spec[2] = 0;
	mtl->exp = 0;
	mtl->rfl = 0;
	mtl->trans = 0;
	mtl->refract = 0;
	mtl->glow = 0;
	mtl->coc = 0;
	mtl->txtlist.ptr = (Material *)NULL;
	mtl->txtlist.next = (Llist *)NULL;

	return(mtl);
}

Cluster	*new_cluster()
{
	Cluster	*cl;

	cl = (Cluster *)malloc(sizeof(Cluster));

	cl->name = (char *)NULL;
	cl->vlist = (int *)NULL;
	cl->ncvs = 0;
	cl->state = 0;

	return(cl);
}

int	new_vertex(me)
Mesh	*me;
{
	Vertex	*v;

	if ( me->maxverts == me->nverts )
	{
		me->maxverts += NUMMALLOC;
		me->vertlist = (Vertex *)realloc( me->vertlist, 
						me->maxverts * sizeof(Vertex) );
	}

	v = &(me->vertlist[me->nverts]);

	v->d[0] = v->d[1] = v->d[2] = 0;
	v->dm = 0;
	v->flag = 0;
	v->xflag = 0;
	v->w = 0;
	v->elist.ptr = (void *)NULL;
	v->elist.next = (Llist *)NULL;
	me->nverts++;

	return(me->nverts-1);
}

int	new_poly(me)
Mesh	*me;
{
	Polygon	*p;

	if ( me->maxpolys == me->npolys )
	{
		me->maxpolys += NUMMALLOC;
		me->polylist = (Polygon *)realloc( me->polylist,
						me->maxpolys * sizeof(Polygon));
	}

	p = &(me->polylist[me->npolys]);
	p->nodes = 0;
	p->mtlnum = 0;
	p->flag = 0;
	p->nodelist = (Node *)NULL;

	me->npolys++;

	return( me->npolys-1 );
}

Node	*new_nodes( num )
int	num;
{
	Node	*n;
	int		i;

	n = (Node *)malloc( sizeof(Node) * num );

	for ( i=0; i<num; i++)
	{
		n[i].vnum = 0;
		n[i].flag = 0;
		n[i].n[0] = 0;	n[i].n[1] = 0;	n[i].n[2] = 0;
		n[i].uv[0] = 0;	n[i].uv[1] = 0;
	}

	return( n );
}
