#include <stdio.h>
#include <math.h>
#include <hrc.h>

void	polynorms( mesh )
Mesh	*mesh;
{
	Vertex	*vlist;
	Node	*nlist;
	Polygon	*plist;
	double	v1[3],v2[3],m1,m2;
	int	vn1,vn2,vn3;
	int	i,j;

	plist = mesh->polylist;
	vlist = mesh->vertlist;

	for (i=0; i<mesh->npolys; i++)
	{
		nlist = plist[i].nodelist;
		vn1 = nlist[0].vnum;
		vn2 = nlist[1].vnum;
		vn3 = nlist[plist[i].nodes-1].vnum;

		for (j=0; j<3; j++)
			v1[j] = vlist[vn2].p[j] - vlist[vn1].p[j];
		
		m1 = 0;
		for (j=0; j<3; j++)
			m1 += ( v1[j] * v1[j] );

		m1 = sqrt ( (double)m1 );
		if ( m1 != 0 )
		{
			for (j=0; j<3; j++)
				v1[j] /= m1;
		}
		else
		{
			for (j=0; j<3; j++)
				v1[j] = 0;
		}

		for (j=0; j<3; j++)
			v2[j] = vlist[vn3].p[j] - vlist[vn1].p[j];
		
		m2 = 0;
		for (j=0; j<3; j++)
			m2 += ( v2[j] * v2[j] );

		m2 = sqrt ( (double)m2 );
		if ( m2 != 0 )
		{
			for (j=0; j<3; j++)
				v2[j] /= m2;
		}
		else
		{
			for (j=0; j<3; j++)
				v2[j] = 0;
		}

		/* cross em' dano */
		plist[i].n[0] = v2[1] * v1[2] - v2[2] * v1[1];
		plist[i].n[1] = v2[2] * v1[0] - v2[0] * v1[2];
		plist[i].n[2] = v2[0] * v1[1] - v2[1] * v1[0];
	}
}

void	pointnorms( mesh )
Mesh	*mesh;
{
	Vertex	*vlist;
	Node	*nlist;
	Polygon	*plist;
	int	i,j;
	float	m;

	plist = mesh->polylist;
	vlist = mesh->vertlist;

	for (i=0; i<mesh->nverts; i++)
	{
		vlist[i].tmp = 0;	/* init nreferences */
		vlist[i].n[0] = 0;	/* init nreferences */
		vlist[i].n[1] = 0;	/* init nreferences */
		vlist[i].n[2] = 0;	/* init nreferences */
	}

	for (i=0; i<mesh->npolys; i++)
	{
		if ((plist[i].n[0] == 0 )
		&& ( plist[i].n[1] == 0 )
		&& ( plist[i].n[2] == 0 ))
			fprintf(stderr,"Face normals == 0 ??\n");

		for ( j=0; j<plist[i].nodes; j++ )
		{
			vlist[plist[i].nodelist[j].vnum].n[0] += plist[i].n[0];
			vlist[plist[i].nodelist[j].vnum].n[1] += plist[i].n[1];
			vlist[plist[i].nodelist[j].vnum].n[2] += plist[i].n[2];
			vlist[plist[i].nodelist[j].vnum].tmp++;	/* Ntimes vertex has been add to */
		}
	}

	for (i=0; i<mesh->nverts; i++)
	{
		if ( vlist[i].tmp != 0 )
		{
			m = sqrt( vlist[i].n[0] * vlist[i].n[0] + 
					vlist[i].n[1] * vlist[i].n[1] + 
					vlist[i].n[2] * vlist[i].n[2] ); 

			vlist[i].n[0] /= m;
			vlist[i].n[1] /= m;
			vlist[i].n[2] /= m;

			/*
			vlist[i].n[0] /= vlist[i].tmp;
			vlist[i].n[1] /= vlist[i].tmp;
			vlist[i].n[2] /= vlist[i].tmp;
			*/
		}
		else
		{
			vlist[i].n[0] = 0;
			vlist[i].n[1] = 0;
			vlist[i].n[2] = 0;
		}
	}

}

buildvtxedgelist( mesh )
Mesh	*mesh;
{
	int	j;

	for (j=0; j<mesh->nedges; j++)
	{
		ins_list( &(mesh->vertlist[mesh->edgelist[j].vptr0].elist),j );
		ins_list( &(mesh->vertlist[mesh->edgelist[j].vptr1].elist),-j );
	}
}

edgelength( mesh )
Mesh	*mesh;
{
	Edge	*el;
	Vertex	*vl;
	int		i;

	el = mesh->edgelist;
	vl = mesh->vertlist;

	for (i=0; i<mesh->nedges; i++)
	{
		el[i].d[0] = vl[ el[i].vptr1 ].p[0] - vl[ el[i].vptr0 ].p[0];
		el[i].d[1] = vl[ el[i].vptr1 ].p[1] - vl[ el[i].vptr0 ].p[1];
		el[i].d[2] = vl[ el[i].vptr1 ].p[2] - vl[ el[i].vptr0 ].p[2];

		el[i].length = sqrt( el[i].d[0] * el[i].d[0] + el[i].d[1] * el[i].d[1] 
			+ el[i].d[2] * el[i].d[2] );
	}
}

flagclstverts( mod )
Model	*mod;
{
	int		i,done;
	Llist	*clstlist;
	Cluster	*clst;
	Vertex	*vlist;

	vlist = mod->mesh->vertlist;

	/* flag verts to xform */
	clstlist = &(mod->clstlist);
	clst = clstlist->ptr;
	if ( clst != (Cluster *)NULL )
	{
		done = 0;
		while ( !done )
		{
			if ( !clst->state )
			{
				for (i=0; i<clst->ncvs; i++)
					vlist[clst->vlist[i]].xflag = 0;
			}

			if ( clstlist->next != (void *)NULL )
			{
				clstlist = clstlist->next;
				clst = clstlist->ptr;
			}
			else
				done = 1;
		}
	}
}

buildedgelist( mesh )
Mesh	*mesh;
{
	int	i,j;
	Polygon	*plist;
	Node	*nlist;

	plist = mesh->polylist;
	for (i=0; i<mesh->npolys; i++)
	{
		nlist = plist[i].nodelist;
		for (j=0; j<(plist[i].nodes-1); j++)
		{
			ins_edge( mesh, nlist[j].vnum, nlist[j+1].vnum );
		}
		ins_edge( mesh, nlist[j].vnum, nlist[0].vnum );
	}
}

ins_edge()
{
}
