#include <stdio.h>
#include <hrc.h>
#include <token.h>

char	*unget_token_ptr,*token_ptr;
char	token_buffer[1048576];
char	atoken[MAXTOKENSIZE];
int		linecounter;

Symtab symbols[] = {
	"PROCESS",		1,
	"AVERAGED",		2<< 1,
	"SELECTED",		2<< 2,
	"STATIC",		2<< 3,
	"INTENSITY",	2<< 4,
	"PHONG",		2<< 5,
	"LAMBERT",		2<< 6,
	"XY",			2<< 7,
	"YX",			2<< 8,
	"XZ",			2<< 9,
	"ZX",			2<<10,
	"YZ",			2<<11,
	"ZY",			2<<12,
	"UV",			2<<13,
	"CARDINAL",		2<<14,
	"HIDDEN",		2<<15,
	"FACETED",		2<<16,
	"HARD",			2<<17,
	"INTENSITY",	2<<18,
	"UV",		2<<19,
	"WITHOUT_REFLECT",		2<<20,
	"",				0,
};

init_tokenizer()
{
	token_ptr = unget_token_ptr = NULL;
	linecounter = 0;
	token_buffer[0] = '\0';
	atoken[0] = '\0';
}

int	mand_token( fp, str, todo )
FILE	*fp;
char	*str;
int		todo;
{
	char	*token;

	token = get_token(fp);
	if ( strcmp(token,str) == 0 )
		return(0);

	/* oops */
	return(notify(str,token,todo));
}

int	opt_token( fp,str )
FILE	*fp;
char	*str;
{
	char	*token;

	token = get_token(fp);
	if ( strcmp(token,str) == 0 )
	{
		unget_token();
		return( 1 );
	}
	else
	{
		unget_token();
		return( 0 );
	}
}

int	mand_symbol(fp,todo)
FILE	*fp;
int		todo;
{
	char	*token;
	int		sym;

	token = get_token(fp);

	if ( (sym=get_symnum(token)) > 0)
		return(sym);
	else
		return(notify("Expecting symbol",token,todo));
}

int	opt_symbol(fp)
FILE	*fp;
{
	char	*token;
	int		sym;

	token = get_token(fp);

	if ( (sym=get_symnum(token)) > 0)
	{
		unget_token();
		return(1);
	}
	else
	{
		unget_token();
		return(0);
	}
}

int	get_symnum( symbol )
char	*symbol;
{
	int	i;

#ifdef DSYM
	fprintf(stderr,"getnum sym=%s\n",symbol);
#endif
	for (i=0; symbols[i].num != 0; i++)
	{
		if ( strcmp( symbols[i].name, symbol ) == 0 )
		{
#ifdef DSYM
	fprintf(stderr,"getnum num=%d\n",symbols[i].num);
#endif
			return( symbols[i].num );
		}
	}
	return(-1);
}

char	*get_symname( num )
int		num;
{
	int	i;

#ifdef DSYM
	fprintf(stderr,"getname sym=%d\n",num);
#endif
	for (i=0; symbols[i].num != 0; i++)
	{
		if ( symbols[i].num & num )
		{
#ifdef DSYM
	fprintf(stderr,"getname lookup=%s\n",symbols[i].name);
#endif
			return( symbols[i].name );
		}
	}
	return("null");
}

char	*get_string(fp)
FILE	*fp;
{
	char	*token,*tmpstr;

	token = get_token(fp);
	tmpstr = (char *)malloc(strlen(token)+1);
	strcpy(tmpstr,token);

	return(tmpstr);
}

int	mand_int( fp, target, str, todo )
FILE	*fp;
int		*target;
char	*str;
int		todo;
{
	char	*token;

	token = get_token(fp);
	if ( sscanf(token,"%d",target) == 1 )
		return(0);

	return( notify( str,token, todo ));
}

int	mand_float( fp, target, str, todo )
FILE	*fp;
float	*target;
char	*str;
int		todo;
{
	char	*token;

	token = get_token(fp);

	if ( strncmp(token,"nan",3) == 0 )
	{
		*target = 0.0;
		return(0);
	}

	if ( sscanf(token,"%f",target) == 1 )
		return(0);

	return( notify( str,token, todo ));
}

int	mand_float2( fp, target, str, todo )
FILE	*fp;
float	target[];
char	*str;
int		todo;
{
	mand_float(fp, &(target[0]),str,todo);
	mand_float(fp, &(target[1]),str,todo);
}

int	mand_float3( fp, target, str, todo )
FILE	*fp;
float	target[];
char	*str;
int		todo;
{
	mand_float(fp, &(target[0]),str,todo);
	mand_float(fp, &(target[1]),str,todo);
	mand_float(fp, &(target[2]),str,todo);
}

int	notify( str, token, todo)
char	*str;
char	*token;
int		todo;
{
	/* oops */
	if ( todo == EXIT )
	{
		fprintf(stderr,"Expecting %s found %s",str,token);
		fprintf(stderr,"At line %d\n",linecounter);
		exit(-1);
	}
	else
	if ( todo == WARN )
	{
		fprintf(stderr,"Warning expecting %s found %s",str,token);
		fprintf(stderr,"At line %d\n",linecounter);
		return(-1);
	}
	else
	if ( todo == QUIET )
	{
		return(-1);
	}
}

char	*get_token(in_fp)
FILE	*in_fp;
{
	int	done,i;

	if ( token_ptr == NULL )
	{
		if ( fgets( token_buffer,1048576,in_fp) == (char *)NULL )
		{
			atoken[0] = '\0';
			return;
		}
		token_ptr = token_buffer;
		linecounter = 0;
	}

	unget_token_ptr = token_ptr;

	while (( *token_ptr == ' ' ) || ( *token_ptr == '\t'))
		 token_ptr++;

	while ( *token_ptr == '\n' )
	{
		if ( fgets( token_buffer,1048576,in_fp) == (char *)NULL )
		{
			atoken[0] = '\0';
			return;
		}
		linecounter++;
		token_ptr = token_buffer;
		while (( *token_ptr == ' ' ) || ( *token_ptr == '\t'))
		 	token_ptr++;
		unget_token_ptr = token_ptr;
	}

	done = FALSE;
	for (i=0;(!done);i++)
	{
		atoken[i] = *token_ptr;
		if (( *token_ptr == '<' ) || ( *token_ptr == '>' )
		||  ( *token_ptr == '[' ) || ( *token_ptr == ']' )
		||  ( *token_ptr == ':' ) || ( *token_ptr == '(' )
		||  ( *token_ptr == ')' ) || ( *token_ptr == ',' )
		||  ( *token_ptr == '{' ) || ( *token_ptr == '}' ))
			done = TRUE;

		token_ptr++;

		if (( *token_ptr == ' ' ) || ( *token_ptr == '\t')
		 || ( *token_ptr == '[' ) || ( *token_ptr == ']' )
		 || ( *token_ptr == '<' ) || ( *token_ptr == '>' )
		 || ( *token_ptr == ':' ) || ( *token_ptr == '(' )
		 || ( *token_ptr == ')' ) || ( *token_ptr == ',' )
		 || ( *token_ptr == '{' ) || ( *token_ptr == '}' )
		 || ( *token_ptr == '\n'))
			done = TRUE;
	}

	atoken[i] = '\0';

/* #define TOKENDEBUG */
#ifdef TOKENDEBUG
	fprintf(stderr,"token [%s]\n",atoken);
#endif
	return(atoken);
}

unget_token()
{
#ifdef TOKENDEBUG
	fprintf(stderr,"unget_token >%s<\n",atoken);
#endif
	token_ptr = unget_token_ptr;
}

flush_comment(in_fp)
FILE	*in_fp;
{
	fgets( token_buffer,1048576,in_fp);
	unget_token_ptr = token_ptr = token_buffer;
	linecounter++;
}

choke(errortext,token)
char	*errortext,*token;
{
	fprintf(stderr,"line %d: %s '%s' found\n",
		linecounter,errortext,token);
	exit(-1);
}
